/*	GaussFit - A System for Least Squares and Robust Estimation

	Source Code Copyright (C) 1987 by William H. Jefferys,
	Michael J. Fitzpatrick and Barbara E. McArthur
	All Rights Reserved.
*/

/* main loop */

#include <math.h>
#include <stdio.h>
#include "array.h"
#include "datum.h"
#include "simpledefs.h"
#include "symboltable.h"
#include "defines.h"
#include "house.h"
#include "gffiles.h"
#include "alloc.h"
#ifdef THINK_C
#include <Memory.h>
#endif
#include "prototypes.p"

#include "util.proto.h"
#include "files.proto.h"
#include "machine.proto.h"
#include "matrixman.proto.h"
#include "declare.proto.h"
#include "symboltable.proto.h"
#include "opcodes.proto.h"
#include "minsum.proto.h"
#include "errhandle.proto.h"
#include "gfprint.proto.h"
#include "house.proto.h"
#include "gauss.proto.h"
#include "xprintf.proto.h"

extern FILE *fp;  /* result file piinter */
/*int totsize, totfree, totrealloc;*/
int running = 0;

double tolerance;  /* tolerance */
double fit=999.0;
double envhuber, envtukey, envfair, envtrim;
int envminsum, envirls, envorm, envprec,envdbg;
double lambda, factor;
int MARQ = 0;
int maxiters;
int space;
int limit;
int iterno;
int tracedisplay = 0;
int showresults = 1;
double *OldPrm = NULL;
double gOldSigma = -1.;
extern int recopyfilecmd;
extern MIDAS *datatable;


void
inititers(void) {
	gOldSigma = -1.;
	space = 0;
}

void
checkdone(void) {
	if(tolerance < fit && iterno>= maxiters) {  /* if reached iters without convergence */
		itlimit(stdout);
		itlimit(fp);
	}
	if (getenvval("prvar") && (fit < tolerance))
		CovarianceMtx();
	if (iterno>= maxiters || fit<tolerance) {
		ChiSquare();
	}
}

#ifndef OBJECTS

void
gaussmain(char *t, char *s)
{
	int i;

	printcopyright(stdout);
	if (recopyfilecmd) {
		system("cp ./untouched/* .");
	}
	
	getenvvars(s);
	xparfileopen1();
	xdatafileopen();

	mycompile(t); /* compile file "t" */
			
	inititers();
	for(i=0;i<maxiters;i++) { /* iterate at most maxite times */
		/* inform user of progress */
		if (oneiter(&gOldSigma)) break;
	}
	if(tolerance < fit)   /* if reached iters without convergence */
	{
		itlimit(stdout);
		itlimit(fp);
	}
	if (getenvval("prvar") && (fit < tolerance))
		CovarianceMtx();

	ChiSquare();
	epilog();  /* clean up */
}

#endif

oneiter(double *oldSigma)
{
	double Curr_Sigma;

	/* initialize Householder routines */
	InitHouse("results", getenvint("prmat"), 10.0);
	xprintf(stdout,"\nStart of Iteration %d...\n\n",iterno+1);
	xprintf(fp,"\nStart of Iteration %d...\n\n",iterno+1);


	Forming_Eqns(); /* form the equations of condition */

	/*if double  iteration method  or Marq thnn proceed with second */
	if ((getenvval("double")) || (MARQ)) {
		datatable = NULL;
		InitHouse("results", getenvint("prmat"), 10.0);
		Forming_Eqns(); /* form the equations of condition */
	}

	if (MARQ) {
		if (!space)
			Allocate_Param_Space();
		
		Curr_Sigma=Get_Sigma();

		if (*oldSigma < 0.00 || Curr_Sigma < *oldSigma) {
			*oldSigma = Curr_Sigma;
			Current_into_Old_Params();
			SolveMatrix(); /* solve the estimation problem */
			ShowResults(iterno+1);/* show results and update parameters */
			lambda = lambda * factor;
		} else {
			Old_into_Current_Params();
			lambda = lambda/(sqr(factor));
			*oldSigma = -1.0;
		}
		xprintf(fp,"lambda = %28.17lf\n",lambda);
		printold();
	} else {
		SolveMatrix(); /* solve the estimation problem */
		ShowResults(iterno+1);/* show results and update parameters */
	}
	
	iterno++;
	if (fit<tolerance) return 1;
	
	return 0;
}

void
Forming_Eqns(void)
{
		xprintf(stdout,"\nForming Eqs...\n\n");
		/* define global variables */
		setpc2label("_Globals");
		running = 1;
		interpret();
		running = 0;
		
		xparfileopen2();	/* fill in fields in declared params */
		
		/* execute user's model starting at main */
		setpc2label("main");
		running = 1;
		interpret();
		running = 0;
		ResetDeclareStack(); /* discard all defined variables */
		SumIt(); /* sum sigmas */
}

double UpdateValues(void)  /* update all parameters */
{
	int i;
	int type;
	double dahat;
	char *name;
	short index[5];
	short symInx, fvInx;
	int indsz,prmsz;
	double error;
	extern double DeltaV;
	double ahat;

    /* cycle through all paramters */
	error = DeltaV; /* error starts with value returned by phi calculation */
	i = 0;

	indsz = getindexsz();
	prmsz = getmaxprmname();
	/* get data on next parameter */
	while(GetDeltaValue(i,&name,index,&type,&dahat))
	{                    
		i++;
		putdelta(name,dahat);
		symInx = findsymbol(name, 0);
		fvInx = FileVarOf(symInx);
		if (filevarptr[fvInx].xvars[4]>0) {
			ahat = getxparval(symInx,index);/* get current value of paramter */
			/* increment value and update paramter file */
			putxparval(symInx,index,ahat + dahat);
		} else {
			ahat = getparamval(symInx); /*get current value of parameter */
			putparamval(symInx,ahat + dahat); /* increment value and update parameter file */
		}
		/* print results in results file and to screen */
		if ((index[4] == 0) && (indsz == 0)) {
			if (showresults)
				xprintf(stdout,"%*s = ",prmsz,name);
			xprintf(fp,"%*s = ",prmsz,name);

		} else if ((index[4] == 0) && (indsz != 0)) {
			if (showresults)
				xprintf(stdout,"%*s%*s = ",prmsz,name,indsz," ");
			xprintf(fp,"%*s = ",prmsz+indsz,name);
		} else {
			if (showresults) {
				xprintf(stdout,"%*s",prmsz, name);
				prIndex(stdout,index,-indsz);
				xprintf(stdout," = ");
			}
			xprintf(fp,"%*s", prmsz,name);
			prIndex(fp,index,-indsz);
			xprintf(fp, " = ");
		}
		if (showresults) {
			printdouble(stdout,ahat+dahat,0);
			xprintf(stdout," delta = ");
			printdouble(stdout,dahat,0);
			xprintf(stdout,"\n");
		}

		printdouble(fp,ahat+dahat,0);
		xprintf(fp," delta = ");
		printdouble(fp,dahat,0);
		xprintf(fp,"\n");

		/* update convergence error */
		error = maxval(fabs(dahat)/(tolerance +fabs(ahat)),error);
	}
	i++;  /* ????? */
	xprintf(stdout,"\n");
	xprintf(fp,"\n");
	return error;
}

void
ShowResults(int count)
{

	PrintResults(); /* print results */
	fit = UpdateValues(); /* update the paramters */
	/* print information for user */
	xprintf(stdout,"fit =\t");
	xprintf(fp,"fit =\t");
	printdouble(fp,fit,0);
	printdouble(stdout,fit,0);
	xprintf(stdout,", tol =\t");
	xprintf(fp,", tol =\t");
	printdouble(stdout,tolerance,0);
	printdouble(fp,tolerance,0);
	xprintf(stdout,"\n");
	xprintf(fp,"\n");
	xprintf(stdout,"\nEnd of Iteration %d...\n\n",count);
	xprintf(fp,"\nEnd of Iteration %d...\n\n",count);
	if (count == 1)
		checkvar();
	if (getenvint("prmwrt") != -1)
		writexpar();
}

void
checkvar(void)  /* check to make sure variances are not a mix of 0 and non-0 */
{
	int z, nz;

	z = getvar_zero();
	nz = getvar_nonzero();
	if ((z == 1) && (nz == 1))
 		fatalerror("FATALERROR! Variances may not be a mix of non-zeros and zeros!\n",
			" ");

}

void
itlimit(FILE *fpp)
{
	xprintf(fpp,"\n***********************************");
	xprintf(fpp,"***********************************\n");
	xprintf(fpp,"\nIteration limit reached.  Convergence has not occurred.\n");
	xprintf(fpp,"\n***********************************");
	xprintf(fpp,"***********************************\n\n");
}

void
prolog(void)
{
}

void
getenvvars(char *s)
{
	openenvtable(s); /* open environment table "s" */
	tolerance = getenvval("tol"); /* get tolerance from environment */
	if ((tolerance < 1.0e-13)   && (tolerance > -1.0e-13))   
		tolerance = 0.000001;
	envprec = getenvint("prec");
	envfair = getenvval("fair");
	envirls = getenvint("irls");
	envorm  = getenvint("orm");
	envtrim = getenvval("trim");
	envdbg  = getenvint("dbg");
	envhuber  = getenvval("huber");
	envtukey  = getenvval("tukey");
	envminsum = getenvint("minsum");
	if (envminsum != 0 && (notzero(envfair)))
		fatalerror("Both minsum and fair have been set to non-zero in the environment file.\n"," ");
	lambda = getenvval("lambda");
	factor = getenvval("factor");
	tracedisplay = getenvint("trace");
	if (notzero(lambda))
		MARQ = 1.0;
	if ((envminsum != 0) && (MARQ))
		fatalerror("Both minsum and lambda have been set to non-zero in the environment file.\nLevenburg-Marquardt cannot be used with minsum.\n"," ");
	if (((lambda < .0000000000001)||(lambda > 100.0)) && (MARQ)) 
	{
		lambda = .0001;
		warningerror("Keyword lambda missing or less than or equal to zero in the environment file.\n"," ",0,0);
	}
	if (((factor < .0000000000001)||(factor >100.0)) && (MARQ)) 
	{
		factor = 0.1;
		warningerror("Keyword factor missing or less than or equal to zero in the environment file.\n"," ",0,0);
	}
	maxiters = getenvint("iters"); /* get max iterations from environment */
	if ((maxiters < 1) || (maxiters >1000))
	{
		maxiters = 10;
		warningerror("Keyword iters missing or less than zero in environment file.\n"," ",0,0);
	}
#ifdef NOTYET
	if (getenvint("derivs") > 1) {
		do2ndDerivs = 1;
		doObservations = 1;
	}
#endif
}


void
Allocate_Param_Space(void)
{
	int i;

	limit = Get_LastCol();
	space++;
	OldPrm = (double*)MemAlloc("OldPrm",(long)(limit+1)*sizeof(double));
	for (i=0;i<=limit;i++)
	{
		*(OldPrm+i) = 0.0;
	}
}

void 
Free_Param_Space() {
	freemem("OldPrm",(char*)OldPrm);
}

void
Old_into_Current_Params(void)
{
	int i, type;
	double dahat,ahat;
	char *name;
	short index[5];
	short symInx, fvInx;


 	i = 0;   /* cycle through all paramters */
	xprintf(fp,"\n");
 	while(GetDeltaParams(i,&name,index,&type,&dahat)) /* get data on next parame
ter */
 	{
		 i++;
		symInx = findsymbol(name, 0);
		fvInx = FileVarOf(symInx);
		if (type != RightHandType) {
			if (filevarptr[fvInx].xvars[4]>0) {
 				ahat = getxparval(symInx,index);/* get current value of paramter */
 				putxparval(symInx,index,OldPrm[i-1]);/* increment value and update paramter file */
			} else {
 				ahat = getparamval(symInx); /*get current value of parameter */
 				putparamval(symInx,OldPrm[i-1]);
 			}
 		}
	}
	xprintf(fp,"\n");
}


void
Current_into_Old_Params(void)
{
	int i, type;
	double dahat;
	char *name;
	short index[5];
	short symInx, fvInx;

	i = 0;   /* cycle through all paramters */
	xprintf(fp,"\n");
	while(GetDeltaParams(i,&name,index,&type,&dahat)) /* get data on next parameter */
	{                    
		i++;
		symInx = findsymbol(name, 0);
		fvInx = FileVarOf(symInx);
		if (type != RightHandType) {
			if (filevarptr[fvInx].xvars[4]>0) {
				OldPrm[i-1] = getxparval(symInx,index);/* get current value of paramter */
			} else {
				OldPrm[i-1] = getparamval(symInx); /*get current value of parameter */
			}
		}
	}

	xprintf(fp,"\n");
}

void
epilog(void)
{
	if (fp != NULL) { fclose(fp); fp = NULL; }
	freeworkspace();
	freerowspace();
	freecolspace();
	freematspace();
	freemeandu();
	freestacks();
	if (MARQ) Free_Param_Space();
	savexpar(); /* save paramter file */
	saveenv(1); /* save environment file */
}

void
printold(void)
{
	int i;
	xprintf(fp,"\n");
	for (i=0;i<=limit;i++)
	{
		xprintf(fp,"OldPrm[%d] = %28.17lf\n",i,OldPrm[i]);
	}
	xprintf(fp,"\n");
}

