#include <time.h>
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include "precision.proto.h"

machine_prec(void)

{

/*
		This program prints hardware-determined double-precision
		machine constants obtained from rmachar.  Dmachar is a C
		translation of the Fortran routine MACHAR from W. J. Cody,
		"MACHAR: A subroutine to dynamically determine machine
		parameters".  TOMS (14), 1988.

		Descriptions of the machine constants are given in the
		prologue comments in rmachar.

		Subprograms called

		  rmachar

		Original driver: Richard Bartels, October 16, 1985

		Modified by: W. J. Cody
						 July 26, 1988

		**********
*/
		double prec;
		int it;

		rmachar(&it);
		prec = ceil(it*0.3);
		return (int)prec;
}


void
rmachar(int *it)

/*

	This subroutine is intended to determine the parameters of the
	 floating-point arithmetic system specified below.  The
	 determination of the first three uses an extension of an algorithm
	 due to M. Malcolm, CACM 15 (1972), pp. 949-951, incorporating some,
	 but not all, of the improvements suggested by M. Gentleman and S.
	 Marovich, CACM 17 (1974), pp. 276-277.  An earlier version of this
	 program was published in the book Software Manual for the
	 Elementary Functions by W. J. Cody and W. Waite, Prentice-Hall,
	 Englewood Cliffs, NJ, 1980.  The present program is a
	 translation of the Fortran 77 program in W. J. Cody, "MACHAR:
	 A subroutine to dynamically determine machine parameters".
	 TOMS (14), 1988.

	Parameter values reported are as follows:

		  it		- the number of base ibeta digits in the floating-point
						significand

		Latest revision - August 4, 1988

		Author - W. J. Cody
					Argonne National Laboratory

*/

{
		int i,irnd;
		int itmp;
		double a,b,beta,one,y,z,zero;
		double betah,t,tmp,tmp1,two;

		irnd = 1;
		one = (double)irnd;
		two = one + one;
		a = two;
		b = a;
		zero = 0.0e0;

/*
  determine beta ala malcolm
*/

		tmp = ((a+one)-a)-one;

		while (tmp == zero) {
			a = a+a;
			tmp = a+one;
			tmp1 = tmp-a;
			tmp = tmp1-one;
		}

		tmp = a+b;
		itmp = (int)(tmp-a);
		while (itmp == 0) {
			b = b+b;
			tmp = a+b;
			itmp = (int)(tmp-a);
		}

		beta = (double)itmp;

/*
  determine  it
*/

		(*it) = 0;
		b = one;
		tmp = ((b+one)-b)-one;

		while (tmp == zero) {
			*it = *it+1;
			b = b*beta;
			tmp = b+one;
			tmp1 = tmp-b;
			tmp = tmp1-one;
		}
}
