
#include <stdio.h>
#include "array.h"
#include "datum.h"
#include "simpledefs.h"
#include "prototypes.p"
#include "opcodes.p"
#include "machine.h"
#include "alloc.h"

#include "xprintf.proto.h"

#include "datum.proto.h"
#include "opcodes.proto.h"
#include "logicops.proto.h"
#include "machine.proto.h"


void
PushTrue(void) {
	long node;
	node = NewNode(0);
	datumptr[node].value = 1.0;
	pushval(node);
}

void
PushFalse(void) {
	long node;
	node = NewNode(0);
	datumptr[node].value = 0.0;
	pushval(node);
}

void
NotFn(void) {
	long in1, in2;
	
	reg[regn++] = in1 = popval();
	trace4list(in1);
	if (datumptr[in1].value == 0.0) PushTrue();
	else PushFalse();
	FreeNodeList(in1);
}


void
AndFn(void) {
	long in1, in2;
	
	reg[regn++] = in2 = popval();
	reg[regn++] = in1 = popval();
	trace4list(in1);
	trace4list(in2);
	if (datumptr[in1].value && datumptr[in2].value) PushTrue();
	else PushFalse();
	FreeNodeList(in1);
	FreeNodeList(in2);
}

void
OrFn(void) {
	long in1, in2;
	
	reg[regn++] = in2 = popval();
	reg[regn++] = in1 = popval();
	trace4list(in1);
	trace4list(in2);
	if (datumptr[in1].value || datumptr[in2].value) PushTrue();
	else PushFalse();
	FreeNodeList(in1);
	FreeNodeList(in2);
}

void
ExclusiveOrFn(void) {
	long in1, in2;
	int val1, val2;
	
	reg[regn++] = in2 = popval();
	reg[regn++] = in1 = popval();
	trace4list(in1);
	trace4list(in2);
	val1 = datumptr[in1].value != 0.0;
	val2 = datumptr[in2].value != 0.0;
	if ((val1 || val2) && !(val1 && val2)) PushTrue();
	else PushFalse();
	FreeNodeList(in1);
	FreeNodeList(in2);
}

void
EqualsFn(void) {
	long in1, in2;
	
	reg[regn++] = in2 = popval();
	reg[regn++] = in1 = popval();
	trace4list(in1);
	trace4list(in2);
	if (datumptr[in1].value == datumptr[in2].value) PushTrue();
	else PushFalse();
	FreeNodeList(in1);
	FreeNodeList(in2);
}

void
NotEqualsFn(void) {
	long in1, in2;
	
	reg[regn++] = in2 = popval();
	reg[regn++] = in1 = popval();
	trace4list(in1);
	trace4list(in2);
	if (datumptr[in1].value != datumptr[in2].value) PushTrue();
	else PushFalse();
	FreeNodeList(in1);
	FreeNodeList(in2);
}

void
LessThanFn(void) {
	long in1, in2;
	
	reg[regn++] = in2 = popval();
	reg[regn++] = in1 = popval();
	trace4list(in1);
	trace4list(in2);
	if (datumptr[in1].value < datumptr[in2].value) PushTrue();
	else PushFalse();
	FreeNodeList(in1);
	FreeNodeList(in2);
}

void
GreaterThanFn(void) {
	long in1, in2;
	
	reg[regn++] = in2 = popval();
	reg[regn++] = in1 = popval();
	trace4list(in1);
	trace4list(in2);
	if (datumptr[in1].value > datumptr[in2].value) PushTrue();
	else PushFalse();
	FreeNodeList(in1);
	FreeNodeList(in2);
}

void
LessOrEqualFn(void) {
	long in1, in2;
	
	reg[regn++] = in2 = popval();
	reg[regn++] = in1 = popval();
	trace4list(in1);
	trace4list(in2);
	if (datumptr[in1].value <= datumptr[in2].value) PushTrue();
	else PushFalse();
	FreeNodeList(in1);
	FreeNodeList(in2);
}

void
GreaterOrEqualFn(void) {
	long in1, in2;
	
	reg[regn++] = in2 = popval();
	reg[regn++] = in1 = popval();
	trace4list(in1);
	trace4list(in2);
	if (datumptr[in1].value >= datumptr[in2].value) PushTrue();
	else PushFalse();
	FreeNodeList(in1);
	FreeNodeList(in2);
}

